/* Copyright (c) James C. Bowman, 1997.
 *
 * With user interface code from Mark J. Kilgard.
 * 'drawCube' and 'drawTorus' by David Yu.
 */

/* This program is freely distributable without licensing fees 
   and is provided without guarantee or warrantee expressed or 
   implied. This program is -not- in the public domain. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>       /* for cos(), sin(), and sqrt() */
#include <assert.h>
#include <GL/glut.h>
#include "trackball.h"

#ifndef M_PI
#define M_PI 3.1415926f
#endif

/************************************************************************/

static GLboolean lighting = GL_TRUE;		/* Enable lighting passes */
static GLboolean specular = GL_TRUE;		/* Enable specular lighting */
static GLboolean drifting = GL_FALSE;		/* Enable texture movement */
static GLboolean linear = GL_FALSE;		/* Enable linear filtering */
static GLboolean useDepth = GL_FALSE;		/* Enable depth buffering */
static GLuint    object = 0;			/* Which object to draw */
static int texWidth = 16, texHeight = 16;	/* Small texture */
static int texChannels = 1;			/* tex color channels 1-4 */
static GLuint octaves = 3;			/* How many octaves of noise */

#define NUM_OBJECTS  2

void
drawCube(void)
{
    glBegin(GL_QUADS);
    glNormal3f(-1.0F, 0.0F, 0.0F);
    glTexCoord2f( 0.0F, 1.0F); glVertex3f(-0.5F,-0.5F,-0.5F);
    glTexCoord2f( 0.0F, 0.0F); glVertex3f(-0.5F,-0.5F, 0.5F);
    glTexCoord2f( 1.0F, 0.0F); glVertex3f(-0.5F, 0.5F, 0.5F);
    glTexCoord2f( 1.0F, 1.0F); glVertex3f(-0.5F, 0.5F,-0.5F);

    glNormal3f( 1.0F, 0.0F, 0.0F);
    glTexCoord2f( 1.0F, 1.0F); glVertex3f( 0.5F, 0.5F, 0.5F);
    glTexCoord2f( 0.0F, 1.0F); glVertex3f( 0.5F,-0.5F, 0.5F);
    glTexCoord2f( 0.0F, 0.0F); glVertex3f( 0.5F,-0.5F,-0.5F);
    glTexCoord2f( 1.0F, 0.0F); glVertex3f( 0.5F, 0.5F,-0.5F);

    glNormal3f( 0.0F,-1.0F, 0.0F);
    glTexCoord2f( 0.0F, 1.0F); glVertex3f(-0.5F,-0.5F,-0.5F);
    glTexCoord2f( 0.0F, 0.0F); glVertex3f( 0.5F,-0.5F,-0.5F);
    glTexCoord2f( 1.0F, 0.0F); glVertex3f( 0.5F,-0.5F, 0.5F);
    glTexCoord2f( 1.0F, 1.0F); glVertex3f(-0.5F,-0.5F, 0.5F);

    glNormal3f( 0.0F, 1.0F, 0.0F);
    glTexCoord2f( 1.0F, 1.0F); glVertex3f( 0.5F, 0.5F, 0.5F);
    glTexCoord2f( 0.0F, 1.0F); glVertex3f( 0.5F, 0.5F,-0.5F);
    glTexCoord2f( 0.0F, 0.0F); glVertex3f(-0.5F, 0.5F,-0.5F);
    glTexCoord2f( 1.0F, 0.0F); glVertex3f(-0.5F, 0.5F, 0.5F);

    glNormal3f( 0.0F, 0.0F,-1.0F);
    glTexCoord2f( 0.0F, 1.0F); glVertex3f(-0.5F,-0.5F,-0.5F);
    glTexCoord2f( 0.0F, 0.0F); glVertex3f(-0.5F, 0.5F,-0.5F);
    glTexCoord2f( 1.0F, 0.0F); glVertex3f( 0.5F, 0.5F,-0.5F);
    glTexCoord2f( 1.0F, 1.0F); glVertex3f( 0.5F,-0.5F,-0.5F);

    glNormal3f( 0.0F, 0.0F, 1.0F);
    glTexCoord2f( 1.0F, 1.0F); glVertex3f( 0.5F, 0.5F, 0.5F);
    glTexCoord2f( 0.0F, 1.0F); glVertex3f(-0.5F, 0.5F, 0.5F);
    glTexCoord2f( 0.0F, 0.0F); glVertex3f(-0.5F,-0.5F, 0.5F);
    glTexCoord2f( 1.0F, 0.0F); glVertex3f( 0.5F,-0.5F, 0.5F);
    glEnd();
}

void
drawTorus(void)
{
    int numMajor = 20;
    int numMinor = 20;
    float majorRadius = 0.6F;
    float minorRadius = 0.2F;
    double majorStep = 2.0F*M_PI / numMajor;
    double minorStep = 2.0F*M_PI / numMinor;
    int i, j;

    for (i=0; i<numMajor; ++i) {
	double a0 = i * majorStep;
	double a1 = a0 + majorStep;
	GLfloat x0 = (GLfloat)cos(a0);
	GLfloat y0 = (GLfloat)sin(a0);
	GLfloat x1 = (GLfloat)cos(a1);
	GLfloat y1 = (GLfloat)sin(a1);

	glBegin(GL_TRIANGLE_STRIP);
	for (j=0; j<=numMinor; ++j) {
	    double b = j * minorStep;
	    GLfloat c = (GLfloat)cos(b);
	    GLfloat r = minorRadius * c + majorRadius;
	    GLfloat z = minorRadius * (GLfloat)sin(b);

	    glNormal3f(x0*c, y0*c, z/minorRadius);
	    glTexCoord2f(i/(GLfloat) numMajor, j/(GLfloat) numMinor);
	    glVertex3f(x0*r, y0*r, z);

	    glNormal3f(x1*c, y1*c, z/minorRadius);
	    glTexCoord2f((i+1)/(GLfloat) numMajor, j/(GLfloat) numMinor);
	    glVertex3f(x1*r, y1*r, z);
	}
	glEnd();
    }
}


/* Apply the given three scalar quantities to the texture matrix,
 * then render the current object.
 */
void
drawObject(GLfloat move,	/* Translation */
	   GLfloat scale,	/* Uniform scale in s and t */
	   GLfloat rot)		/* Rotation around (0, 0) */
{
    glMatrixMode(GL_TEXTURE);
    glPushMatrix();

    glScalef(scale, scale, 1.0f);
    glRotatef(rot, 0.0f, 0.0f, 1.0f);
    glTranslatef(move, 0.0f, 0.0f);

    if (object == 0)
	drawCube();
    else
	drawTorus();

    glPopMatrix();
}

static void
multi(void)
{
    static GLfloat white[] = {1.0f, 1.0f, 1.0f, 1.0f};
    static GLfloat black[] = {0.0f, 0.0f, 0.0f, 1.0f};
    GLuint pass;
    GLfloat freq, amplitude;
    static GLfloat drift;

    srand(9467); /* We need the same sequence of random numbers every
		  * time */

    /*
     * PASS     Blending   Lighting Shade    DepthFunc  DepthMask
     *
     * tex 0	   No	     No     FLAT      LEQUAL       True
     * tex 1      Yes[1]     No     FLAT      EQUAL        False
     * tex n      Yes[1]     No     FLAT      EQUAL        False
     * diffuse    Yes[2]     Yes    SMOOTH    EQUAL        False
     * specular   Yes[1]     Yes    SMOOTH    EQUAL        False
     *
     * Notes:
     *   [1]   glBlendFunc(GL_ONE, GL_ONE)
     *   [2]   glBlendFunc(GL_ZERO, GL_SRC_COLOR)
     */

    if (useDepth) {
      glEnable(GL_DEPTH_TEST);
    } else {
      glDisable(GL_DEPTH_TEST);
    }

    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glDisable(GL_LIGHTING);
    glDisable(GL_DITHER);
    glShadeModel(GL_FLAT);

    if (useDepth) {
	glDepthFunc(GL_LEQUAL);
	glDepthMask(GL_TRUE);
    }

    if (octaves == 0) {
	/* No texture, just draw a solid white cube */

	glDisable(GL_BLEND);
	glDisable(GL_TEXTURE_2D);

	glColor3f(1.0f, 1.0f, 1.0f);
	drawObject(drift, 1.0f, 0.0f);
    } else {
	/* Build up the texture in several passes.  See
	 * "Texturing and Modeling - A Procedural Approach"
	 * (ISBN 0-12-228760-6) page 86.
	 */
	glEnable(GL_TEXTURE_2D);

	freq = 1.0f / 1.0f;
	amplitude = 0.5F;

	for (pass = 0; pass < octaves; pass++) {
	    if (pass == 0) {
		glDisable(GL_BLEND);
	    } else {
		glEnable(GL_BLEND);
		glBlendFunc(GL_ONE, GL_ONE);
		if (useDepth) {
		    glDepthFunc(GL_EQUAL);
		    glDepthMask(GL_FALSE);
		}
	    }

	    /* Because we're using GL_MODULATE, this scales the
	     * contribution from this pass.
	     */
	    glColor3f(amplitude, amplitude, amplitude);

	    /* Use a random orientation for each texture to hide
	     * the underlying square lattice (page 102).
	     */
	    drawObject(drift, freq, (GLfloat)(rand() % 360));

	    /* Changing these two factors can have interesting effects... */
	    freq *= 2.0f;	/* .. this is the lacunarity (page 89) */
	    amplitude *= 0.5f;  /* .. increasing this gives a grainier feel. */
	}
    }

    /* Now light the cube.  */
    if (lighting) {
	glEnable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glShadeModel(GL_SMOOTH);

	if (useDepth) {
	    glDepthFunc(GL_EQUAL);
	    glDepthMask(GL_FALSE);
	}

	/* First, diffuse illumination.  The blend func is
	 * (GL_ZERO, GL_SRC_COLOR) because we want to modulate
	 * the texture by the lighting color.
	 * This pass only needs the diffuse component
	 * of any lighting calculation, hence the material
	 * setup.
	 */

	/* If this is the final pass, we should dither */
	if (!specular)
	    glEnable(GL_DITHER);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, white);
	glMaterialfv(GL_FRONT, GL_SPECULAR, black);
	glBlendFunc(GL_ZERO, GL_SRC_COLOR);
	drawObject(drift, 1.0f, 0.0f);

	/* Second, specular component, which is additive,
	 * so use blend func (GL_ONE, GL_ONE).
	 * Note that the material this time is set up
	 * to give _just_ the specular component of lighting.
	 */
	if (specular) {
	    glEnable(GL_DITHER);
	    /* This must be the final pass, so enable dither */
	    glMaterialfv(GL_FRONT, GL_DIFFUSE, black);
	    glMaterialfv(GL_FRONT, GL_SPECULAR, white);
	    glMaterialf(GL_FRONT, GL_SHININESS, 20.0f);

	    glBlendFunc(GL_ONE, GL_ONE);
	    drawObject(drift, 1.0f, 0.0f);
	}
    }

    if (drifting)
	drift += 0.001f;
}

/* Create and download the application texture map.
 * It's just filled with random texels.
 */

static void
setRandomTexture(void)
{
    int texSize;
    void *textureBuf;
    GLubyte *p;
    int i,j;

    texSize = texWidth*texHeight*texChannels;

    textureBuf = malloc(texSize);
    if (NULL == textureBuf) return;

    p = (GLubyte *)textureBuf;
    for (i=0; i < texWidth; i++) {
	for (j=0; j < texHeight; j++) {
	    int c;

	    for (c = 0; c < texChannels; c++)
		*p++ = rand() & 0xff;
	}
    }

    gluBuild2DMipmaps(GL_TEXTURE_2D, texChannels, texWidth, texHeight, 
		      (texChannels == 1) ? GL_LUMINANCE :
		      (texChannels == 2) ? GL_LUMINANCE_ALPHA :
		      (texChannels == 3) ? GL_RGB :
		      GL_RGBA,
		      GL_UNSIGNED_BYTE, textureBuf);
    free(textureBuf);
}

/************************************************************************
    Code for here on is taken from Mark J. Kilgard's 'dinospin.c'
************************************************************************/

GLfloat angle = -150.0f;   /* in degrees */
GLboolean doubleBuffer = GL_TRUE, iconic = GL_FALSE, keepAspect = GL_FALSE;
int spinning = 0, moving = 0;
int beginx, beginy;
int W = 300, H = 300;
float curquat[4];
GLfloat bodyWidth = 2.0f;
int newModel = 1;
int scaling;
float scalefactor = 1.0f;
float lastquat[4];

GLfloat lightZeroPosition[] = {10.0f, 4.0f, 10.0f, 1.0f};
GLfloat lightZeroColor[] = {0.8f, 1.0f, 0.8f, 1.0f}; /* green-tinted */
GLfloat lightOnePosition[] = {-1.0f, -2.0f, 1.0f, 0.0f};
GLfloat lightOneColor[] = {0.6f, 0.3f, 0.2f, 1.0f}; /* red-tinted */
GLfloat skinColor[] = {0.1f, 1.0f, 0.1f, 1.0f}, eyeColor[] = {1.0f, 0.2f, 0.2f, 1.0f};

static void
recalcModelView(void)
{
  GLfloat m[4][4];

  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
  glPushMatrix();
  build_rotmatrix(m, curquat);
  glMultMatrixf(&m[0][0]);
  glScalef(scalefactor, scalefactor, scalefactor);

  newModel = 0;
}

void
redraw(void)
{
  if (newModel)
    recalcModelView();

  if (useDepth)
      glDepthMask(GL_TRUE);

  glClear(GL_COLOR_BUFFER_BIT | (useDepth ? GL_DEPTH_BUFFER_BIT : 0));

  if (linear) {
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  } else {
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  }

  multi();

  glutSwapBuffers();
}

void
myReshape(int w, int h)
{
  glViewport(0, 0, w, h);
  W = w;
  H = h;
}

void
mouse(int button, int state, int x, int y)
{
  if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
    spinning = 0;
    moving = 1;
    trackball(lastquat, 0.0f, 0.0f, 0.0f, 0.0f);
    if (!spinning && !drifting)
	glutIdleFunc(NULL);
    beginx = x;
    beginy = y;
    if(glutGetModifiers() & GLUT_ACTIVE_SHIFT) {
      scaling = 1;
    } else {
      scaling = 0;
    }
  }
  if (button == GLUT_LEFT_BUTTON && state == GLUT_UP) {
    moving = 0;
  }
}

void
animate(void)
{
  add_quats(lastquat, curquat, curquat);
  newModel = 1;
  glutPostRedisplay();
}

void
motion(int x, int y)
{
  if (scaling) {
    scalefactor = scalefactor * (1.0f + (((float) (beginy - y)) / H));
    beginx = x;
    beginy = y;
    newModel = 1;
    glutPostRedisplay();
    return;
  }
  if (moving) {
    trackball(lastquat,
      (2.0f*beginx - W) / W,
      (H - 2.0f*beginy) / H,
      (2.0f*x - W) / W,
      (H - 2.0f*y) / H
      );
    beginx = x;
    beginy = y;
    spinning = 1;
    glutIdleFunc(animate);
  }
}

GLboolean lightZeroSwitch = GL_TRUE, lightOneSwitch = GL_TRUE;

void
controlLights(int value)
{
  switch (value) {
  case 1:
    lightZeroSwitch = !lightZeroSwitch;
    if (lightZeroSwitch) {
      glEnable(GL_LIGHT0);
    } else {
      glDisable(GL_LIGHT0);
    }
    break;
  case 2:
    lightOneSwitch = !lightOneSwitch;
    if (lightOneSwitch) {
      glEnable(GL_LIGHT1);
    } else {
      glDisable(GL_LIGHT1);
    }
    break;
  case 4:
    glutFullScreen();
    break;
  case 5:
    exit(0);
    break;
  }
  glutPostRedisplay();
}

void
vis(int visible)
{
  if (visible == GLUT_VISIBLE) {
    if (spinning || drifting)
      glutIdleFunc(animate);
  } else {
    if (spinning)
      glutIdleFunc(NULL);
  }
}


/* Dump an image of the screen in a ppm file, numbered sequentially
 * image000.ppm, image001.ppm etc.
 */
static void
image(void)
{
    static GLuint serial = 0;
    char name[10];
    FILE *snap;
    unsigned char *screen_image;

    sprintf(name,
	    "image%03d.ppm",
	    serial++);
    snap = fopen(name, "wb");
    
    if (snap != NULL) {
	/* Write the PPM header */
	fprintf(snap, "P6\n%d %d\n255\n", W, H);

	screen_image = calloc(W * H, 3);
	if (screen_image) {
	    /* Read the entire screen */
	    glReadPixels(0, 0, W, H, GL_RGB, GL_UNSIGNED_BYTE, screen_image);

	    /* Just send the whole image to the PPM file */
	    fwrite(screen_image, W * H, 3, snap);
	    free(screen_image);
	}
	fclose(snap);
    }
}

void Key(unsigned char key, int x, int y)
{
    switch (key) {
    case 27:
	printf("\n\n");
	exit(0);

    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
	octaves = key - '0';
	glutPostRedisplay();
	break;

    case 'd':
	drifting = !drifting;
	if (!spinning)
	    trackball(lastquat, 0.0f, 0.0f, 0.0f, 0.0f);
	glutIdleFunc(animate);
	break;
    case 'f':
	linear = !linear;
	glutPostRedisplay();
	break;
    case 'i':
	image();
	break;
    case 'l':
	lighting = !lighting;
	glutPostRedisplay();
	break;
    case 'o':
	object = (object + 1) % NUM_OBJECTS;
	glutPostRedisplay();
	break;
    case 's':
	specular = !specular;
	glutPostRedisplay();
	break;
    case 'z':
	useDepth = !useDepth;
	glutPostRedisplay();
	break;
    }
}


int
main(int argc, char **argv)
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
  trackball(curquat, 0.0f, 0.0f, 0.0f, 0.0f);
  glutCreateWindow("Multi texture");
  glutDisplayFunc(redraw);
  glutReshapeFunc(myReshape);
  glutVisibilityFunc(vis);
  glutMouseFunc(mouse);
  glutKeyboardFunc(Key);
  glutMotionFunc(motion);
  glutCreateMenu(controlLights);
  glutAddMenuEntry("Toggle right light", 1);
  glutAddMenuEntry("Toggle left light", 2);
  glutAddMenuEntry("Full screen", 4);
  glutAddMenuEntry("Quit", 5);
  glutAttachMenu(GLUT_RIGHT_BUTTON);

  setRandomTexture();

  glClearColor(0.0f, 0.0f, 0.0f, 1.0F);
  glEnable(GL_CULL_FACE);

  glEnable(GL_LIGHTING);
  glMatrixMode(GL_PROJECTION);
  gluPerspective( /* field of view in degree */ 40.0f,
                  /* aspect ratio             */ 1.0f,
                  /* Z near */ 0.1f,
		  /* Z far */ 40.0f);
  glMatrixMode(GL_MODELVIEW);
  gluLookAt(0.0f, 0.0f, 3.0f,  /* eye is at (0,0,30) */
    0.0f, 0.0f, 0.0f,      /* center is at (0,0,0) */
    0.0f, 1.0f, 0.0f);      /* up is in positive Y direction */
  glEnable(GL_NORMALIZE); /* In case we're scaling */

  glPushMatrix();       /* dummy push so we can pop on model
                           recalc */

  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
  glLightfv(GL_LIGHT0, GL_POSITION, lightZeroPosition);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, lightZeroColor);
  glLightfv(GL_LIGHT0, GL_SPECULAR, lightZeroColor);
  glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 0.1f);
  glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.05f);
  glLightfv(GL_LIGHT1, GL_POSITION, lightOnePosition);
  glLightfv(GL_LIGHT1, GL_DIFFUSE, lightOneColor);
  glLightfv(GL_LIGHT1, GL_SPECULAR, lightOneColor);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHT1);
  glLineWidth(2.0f);
  glutMainLoop();
  return 0;             /* ANSI C requires main to return int. */
}
